/* Kernel module help for Alpha.
   Copyright (C) 2002  Richard Henderson.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */

#include <linux/moduleloader.h>
#include <linux/elf.h>
#include <linux/vmalloc.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/kernel.h>

#if 1
#define DEBUGP printk
#else
#define DEBUGP(fmt , ...)
#endif

void *module_alloc(Elf_Ehdr *hdr, Elf_Dyn *dyn, unsigned long size,
		   unsigned long *extra)
{
	return (vmalloc(size) ?: ERR_PTR(-ENOMEM));
}

void
module_free(void *region)
{
	vfree(region);
}

/* Apply the given relocations.  Return -error or 0. */
int apply_relocate(Elf_Rel *rel,
		   unsigned int num_rel,
		   Elf_Sym *syms,
		   const char *strtab,
		   struct module *me)
{
	return -ENOEXEC;
}

/* Apply the given add relocations.  Return -error or 0 */
int apply_relocate_add(Elf_Rela *rela,
		       unsigned int num_rela,
		       Elf_Sym *syms,
		       const char *strtab,
		       struct module *me)
{
	unsigned long i;

	DEBUGP("Doing %u relocations\n", num_rela);
	for (i = 0; i < num_rela; i++) {
		unsigned long r_type = ELF64_R_TYPE(rela[i].r_info);
		char *location = (void *)me->loadaddr + rela[i].r_offset;
		unsigned int symnum;
		unsigned long value = 0;

		if (r_type == R_ALPHA_RELATIVE) {
			/* Binutils before 2.12 or so are borken.  We should
			   have the RELATIVE offset as the addend of the 
			   relocation.  If it's not present, fall back to the
			   value at the relocation address.  */
			u64 addend = rela[i].r_addend;
			if (addend == 0)
				addend = *(u64 *)location;

			*(u64 *)location = (u64)me->loadaddr + addend;
			continue;
		}

		/* This is the symbol it is referring to.  */
		symnum = ELF64_R_SYM(rela[i].r_info);
		if (symnum) {
			value = (u64) resolve_symbol(me, symnum, syms, strtab,
						     NULL);
			if (!value)
				  return -ENOENT;
		}
		value += rela[i].r_addend;

		switch (r_type) {
		case R_ALPHA_GLOB_DAT:
		case R_ALPHA_JMP_SLOT:
		case R_ALPHA_REFQUAD:
			*(u64 *)location = value;
			break;
		default:
			printk(KERN_ERR "module %s: Unknown relocation: %lu\n",
			       me->name, r_type);
			return -ENOEXEC;
		}
	}

	return 0;
}

int
module_finalize(Elf_Ehdr *hdr, struct module *me)
{
	return 0;
}

void module_truncate(struct module *me)
{
	/* FIXME: Implement */
}
