/*  Kernel module help for m68k.
    Copyright (C) 2001 Rusty Russell.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <linux/moduleloader.h>
#include <linux/elf.h>
#include <linux/vmalloc.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/kernel.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(fmt...)
#endif

void *module_alloc(Elf_Ehdr *hdr, Elf_Dyn *dyn, unsigned long modsize,
		   unsigned long *extra)
{
	return vmalloc(modsize) ?: ERR_PTR(-ENOMEM);
}

void module_free(void *modmem)
{
	vfree(modmem);
}

int apply_relocate(Elf_Rel *rel,
		   unsigned int num_rel,
		   Elf_Dyn *dyn,
		   Elf_Sym *syms,
		   const char *strtab,
		   struct module *me)
{
	printk(KERN_ERR "module %s: non-ADD RELOCATION unsupported\n",
	       me->name);
	return -ENOEXEC;
}

/* Apply the given add relocations.  Return -error or 0 */
int apply_relocate_add(Elf_Rela *rela,
		       unsigned int num_rela,
		       Elf_Dyn *dyn,
		       Elf_Sym *syms,
		       const char *strtab,
		       struct module *me)
{
	unsigned int i;
	u32 *location;
	u32 value;

	printk("Applying ADD relocations\n");
	for (i = 0; i < num_rela; i++) {
		unsigned int symnum;
		
		/* This is where to make the change */
		location = me->loadaddr + rela[i].r_offset;

		/* This is the symbol it is referring to */
		symnum = ELF32_R_SYM(rela[i].r_info);
		if (symnum) {
			void *addr = resolve_symbol(me, symnum, syms, strtab,
						    NULL);
			if (!addr)
				return -ENOEXEC;
			value = (unsigned long)addr;
		} else
			value = 0;

		/* Everything is relative */
		value += rela[i].r_addend;

		switch (ELF32_R_TYPE(rela[i].r_info)) {
		case R_68K_RELATIVE:
			*location += (u32)me->loadaddr;
			break;

		case R_68K_PC32:
			*location = value - (u32)location;
			break;

		case R_68K_32:
			*location = value;
			break;

		default:
			printk("Unknown ADD relocation: %u\n",
			       ELF32_R_TYPE(rela[i].r_info));
			return -ENOEXEC;
		}
	}

	return 0;
}

int module_finalize(Elf_Ehdr *hdr, struct module *me)
{
	return 0;
}

void module_truncate(struct module *me)
{
}
