/* Hodgepodge of fake kernel functions. */
#ifndef _KERN_SUPPORT_H
#define _KERN_SUPPORT_H
#include <stddef.h>
#include <string.h>
#include <fcntl.h>

#define KERN_ERR ""
#define KERN_WARNING ""
#define BUG() abort()
#define BUG_ON(cond) do { if (cond) abort(); } while(0)
#define ALIGN(x,a) (((x)+(a)-1)&~((a)-1))
/* Arch dependent, actually */
#define SMP_CACHE_BYTES 64
#define ____cacheline_aligned __attribute__((__aligned__(SMP_CACHE_BYTES)))

#define __stringify_1(x)	#x
#define __stringify(x)		__stringify_1(x)

#define __deprecated

#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

typedef unsigned long long u64;
typedef long long s64;
typedef unsigned int u32;
typedef int s32;
typedef unsigned short u16;
typedef short s16;
typedef unsigned char u8;
typedef signed char s8;

typedef unsigned long long __u64;
typedef long long __s64;
typedef unsigned int __u32;
typedef int __s32;
typedef unsigned short __u16;
typedef short __s16;
typedef unsigned char __u8;
typedef signed char __s8;

typedef unsigned int uint32_t;
typedef unsigned short uint16_t;

static inline void prefetch(const void *x) {;}
#define wmb()
#define asmlinkage

#undef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)

#define NR_CPUS 1
#define get_cpu() 0
#define put_cpu()
#define likely(x) (x)
#define unlikely(x) (x)
#define spin_lock_irq(lock) (*(lock) = 1)
#define spin_unlock_irq(lock) (*(lock) = 0)
#define spin_lock_irqsave(lock, flags) (*(lock) = 1, (flags) = 1)
#define spin_unlock_irqrestore(lock, flags) (*(lock) = 0, (flags) = 0)
typedef int spinlock_t;
#define SPIN_LOCK_UNLOCKED 0

#define DECLARE_MUTEX(mutex) unsigned int mutex
#define down_interruptible(mutex) ((*(mutex) = 1), 0)
#define down(mutex) ((*mutex) = 1)
#define up(mutex) (*(mutex) = 0)

#define elf_check_arch(x) 1

/**
 * container_of - cast a member of a structure out to the containing structure
 *
 * @ptr:	the pointer to the member.
 * @type:	the type of the container struct this is embedded in.
 * @member:	the name of the member within the struct.
 *
 */
#define container_of(ptr, type, member) ({			\
        const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
        (type *)( (char *)__mptr - offsetof(type,member) );})

struct exception_table_entry
{
	unsigned long insn, fixup;
};

typedef struct 
{
	unsigned val;
} atomic_t;

#define atomic_inc(x) ((x)->val++)
#define atomic_dec(x) ((x)->val--)
#define atomic_set(x,v) ((x)->val = v)
#define atomic_read(x) ((x)->val)

struct task_struct
{
	int state;
};

extern struct task_struct *current;
#define TASK_RUNNING 1
#define TASK_UNINTERRUPTIBLE 2
#define set_current_state(s) ((current)->state = s)
#define schedule()
#define synchronize_kernel()

struct seq_file;
#define seq_printf(m, fmt, args...)
struct seq_operations
{
	void *start, *next, *stop, *show;
};

void flush_icache_range(unsigned long start, unsigned long end);
#define local_irq_disable()
#define local_irq_enable()
#define wake_up_process(proc) BUG_ON(proc != current)
#define capable(x) 1

/* Prototypes */
void abort(void);
int sprintf(char *str, const char *format, ...);
int printf(const char *format, ...);
void *malloc(size_t size);
void *free(void *ptr);
void *memcpy(void *dest, const void *src, size_t n);
unsigned long int strtoul(const char *nptr, char **endptr, int base);
long int strtol(const char *nptr, char **endptr, int base);

#define printk printf

#define kmalloc(x,f) malloc(x)
#define kfree(x) free(x)

#define vmalloc(x) malloc(x)
#define vfree(x) free(x)

#define strncpy_from_user(dst, src, count) (strncpy((dst), (src), (count)), 0)
#define copy_from_user(p,up,n) ({ memcpy((p),(up),(n)); 0; })
#define strlen_user(s) strlen(s)

#define simple_strtoul(v, ep, b) strtoul((v), (ep), (b))
#define simple_strtol(v, ep, b) strtol((v), (ep), (b))

#define	ENOSYS		38	/* Function not implemented */
#define	ENOENT		 2	/* No such file or directory */
#define	ENOSPC		28	/* No space left on device */
#define	EINVAL		22	/* Invalid argument */
#define	EPERM		 1	/* Operation not permitted */
#define	EFAULT		14	/* Bad address */
#define	EINTR		 4	/* Interrupted system call */
#define	EBUSY		16	/* Device or resource busy */
#define	EWOULDBLOCK	11	/* Try again */
#define	ENOMEM		12	/* Out of memory */
#define	ENOEXEC		 8	/* Exec format error */
#define	EEXIST		17	/* File exists */
#undef NULL
#define NULL 0

#endif
