#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include "kernsupport.h"
#include <asm/unistd.h>
#include <linux/moduleloader.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

extern int __initfn(void);
extern void __exitfn(void);
extern struct list_head modules;
extern void __init extable_init(void);

static void test_string(const char *str)
{
	printk("Hi: \"%s\"\n",str);
}

EXPORT_SYMBOL(printf);
EXPORT_SYMBOL(test_string);

struct exception_table_entry __start___ex_table[0];
struct exception_table_entry __stop___ex_table[0];

/* Fake current */
struct task_struct __current, *current = &__current;

static void illinst(int signal, siginfo_t *info, void *unused)
{
	char string[100];

	fflush(stdout);
	sprintf(string, "Signal %s, addr %p\n",
		signal == SIGILL ? "SIGILL" :
		signal == SIGSEGV ? "SIGSEGV" :
		signal == SIGTRAP ? "SIGTRAP" :
		"UNKNOWN", info->si_addr);
	write(2, string, strlen(string));
}

asmlinkage long
sys_init_module(void *umod, size_t len, const char *uargs);

int main(int argc, char *argv[])
{
	int fd;
	struct stat st;
	void *map;
	long err;
	struct sigaction act;
	struct module *module;

	memset(&act, 0, sizeof(act));
	act.sa_sigaction = illinst;
	act.sa_flags = (SA_ONESHOT|SA_SIGINFO);

#if 1
	sigaction(SIGILL, &act, NULL);
	sigaction(SIGTRAP, &act, NULL);
	sigaction(SIGSEGV, &act, NULL);
#endif

	if (argc < 2) {
		fprintf(stderr, "usage: %s filename [args]\n",
			argv[0]);
		exit(1);
	}

	/* Initialize the symbol and exception tables */
	extable_init();

	fd = open(argv[1], O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "can't open `%s'\n", argv[1]);
		exit(1);
	}

	fstat(fd, &st);
	map = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
	err = sys_init_module(map, st.st_size, argv[2] ?: "");
	
	if (err < 0) {
		printf("Error %li (%s)\n", -err, strerror(-err));
		return 1;
	}

	module = list_entry(modules.next, struct module, list);
	printf("Module %p\n", module);
#if 0
	printf("(Exported %u syms)\n",
	       module->symbols.end - module->symbols.start);
	printf("init is at %p\n", module->init);

	{
		const struct kernel_symbol *i;

		for (i = module->symbols.start; i < module->symbols.end; i++) {
			printk("EXPORTED: %s = %p\n",
			       i->name, (void *)i->value);
		}
	}

	printf("init = %i\n", module->init());
	printf("module->cut_here = %p (of %p - %p)\n",
	       module->cut_here, module->loadaddr,
	       module->loadaddr + module->module_size);
#endif

	return 0;
}
